#!/usr/bin/env python

"""Tests for `spade_pubsub` package."""
import pytest

from uuid import uuid4
from spade.behaviour import OneShotBehaviour
from .factories import PubSubAgentFactory


XMPP_SERVER = "localhost"
AGENT_DOMAIN = 'localhost'

AGENT_JID = f"pubsuba@{AGENT_DOMAIN}"
AGENT_JID_2 = f"pubsubb@{AGENT_DOMAIN}"
PUBSUB_JID = f"pubsub.{XMPP_SERVER}"
TEST_NODE = "Test_Node"
ITEM_ID = str(uuid4())
TEST_PAYLOAD = "Testing"
TEST_PAYLOAD2 = "Testing2"


@pytest.mark.asyncio
async def test_delete_node(server):
    class DeleteNodeBehaviour(OneShotBehaviour):
        async def run(self):
            self.agent.client.register_plugin('xep_0004')
            config_form = self.agent.client.plugin["xep_0004"].make_form(ftype="submit")
            config_form.addField('pubsub#persist_items', value=True)

            try:
                await self.agent.pubsub.create(PUBSUB_JID, TEST_NODE, config_form)
            except Exception as e:
                print(f"{e}: Node already existed.")
            await self.agent.pubsub.delete(PUBSUB_JID, TEST_NODE)
            result = await self.agent.pubsub.get_nodes(PUBSUB_JID)
            self.kill(exit_code=result)

    agent = PubSubAgentFactory(jid=AGENT_JID)

    await agent.start(auto_register=True)
    assert agent.is_alive() is True

    behaviour = DeleteNodeBehaviour()
    agent.add_behaviour(behaviour)
    await behaviour.join()

    assert len(behaviour.exit_code) == 0

    await agent.stop()
    assert agent.is_alive() is False


@pytest.mark.asyncio
async def test_create_node(server):
    agent = PubSubAgentFactory(jid=AGENT_JID)

    await agent.start(auto_register=True)
    assert agent.is_alive() is True

    class CreateNodeBehaviour(OneShotBehaviour):
        async def run(self):
            await self.agent.pubsub.create(PUBSUB_JID, TEST_NODE)
            result = await self.agent.pubsub.get_nodes(PUBSUB_JID)
            await self.agent.pubsub.delete(PUBSUB_JID, TEST_NODE)
            self.kill(exit_code=result)

    behaviour = CreateNodeBehaviour()
    agent.add_behaviour(behaviour)
    await behaviour.join()

    assert len(behaviour.exit_code) > 0

    node = behaviour.exit_code[0]
    assert node['jid'] == 'pubsub.localhost'
    assert node['node'] == 'Test_Node'
    assert node['name'] == 'leaf'

    await agent.stop()
    assert agent.is_alive() is False


@pytest.mark.asyncio
async def test_purge_node(server):
    agent = PubSubAgentFactory(jid=AGENT_JID)

    await agent.start(auto_register=True)
    assert agent.is_alive() is True

    agent.client.register_plugin('xep_0004')
    config_form = agent.client.plugin["xep_0004"].make_form(ftype="submit")
    config_form.addField('pubsub#persist_items', value=True)

    class PurgeNodeBehaviour(OneShotBehaviour):
        async def run(self):
            await self.agent.pubsub.create(PUBSUB_JID, TEST_NODE, config_form)
            await self.agent.pubsub.publish(PUBSUB_JID, TEST_NODE, TEST_PAYLOAD)
            await self.agent.pubsub.publish(PUBSUB_JID, TEST_NODE, TEST_PAYLOAD2)
            result1 = await self.agent.pubsub.get_items(PUBSUB_JID, TEST_NODE)
            await self.agent.pubsub.purge(PUBSUB_JID, TEST_NODE )
            result2 = await self.agent.pubsub.get_items(PUBSUB_JID, TEST_NODE)
            await self.agent.pubsub.delete(PUBSUB_JID, TEST_NODE)
            self.kill(exit_code=(result1, result2))

    behaviour = PurgeNodeBehaviour()
    agent.add_behaviour(behaviour)
    await behaviour.join()

    before, after = behaviour.exit_code
    assert len(before) == 2
    assert TEST_PAYLOAD in [i[1].text for i in before]
    assert TEST_PAYLOAD2 in [i[1].text for i in before]
    assert len(after) == 0

    await agent.stop()
    assert agent.is_alive() is False


@pytest.mark.asyncio
async def test_subscribe_to_node(server):
    agent = PubSubAgentFactory(jid=AGENT_JID)

    await agent.start(auto_register=True)
    assert agent.is_alive() is True

    class SubscribeBehaviour(OneShotBehaviour):
        async def run(self):
            await self.agent.pubsub.create(PUBSUB_JID, TEST_NODE)
            await self.agent.pubsub.subscribe(PUBSUB_JID, TEST_NODE)
            result = await self.agent.pubsub.get_node_subscriptions(
                target_jid=PUBSUB_JID, target_node=TEST_NODE
            )
            await self.agent.pubsub.delete(PUBSUB_JID, TEST_NODE)
            self.kill(exit_code=result)

    behaviour = SubscribeBehaviour()
    agent.add_behaviour(behaviour)
    await behaviour.join()

    assert len(behaviour.exit_code) == 1
    assert behaviour.exit_code[0] == AGENT_JID

    await agent.stop()
    assert agent.is_alive() is False


@pytest.mark.asyncio
async def test_unsubscribe_from_node(server):
    agent = PubSubAgentFactory(jid=AGENT_JID)

    await agent.start(auto_register=True)
    assert agent.is_alive() is True

    class SubscribeBehaviour(OneShotBehaviour):
        async def run(self):
            await self.agent.pubsub.create(PUBSUB_JID, TEST_NODE)
            subid = await self.agent.pubsub.subscribe(PUBSUB_JID, TEST_NODE)
            await self.agent.pubsub.unsubscribe(PUBSUB_JID, TEST_NODE, AGENT_JID, subid)
            result = await self.agent.pubsub.get_node_subscriptions(
                PUBSUB_JID, TEST_NODE
            )
            await self.agent.pubsub.delete(PUBSUB_JID, TEST_NODE)
            self.kill(exit_code=result)

    behaviour = SubscribeBehaviour()
    agent.add_behaviour(behaviour)
    await behaviour.join()

    assert len(behaviour.exit_code) == 0

    await agent.stop()
    assert agent.is_alive() is False


@pytest.mark.asyncio
async def test_notify(server):
    agent = PubSubAgentFactory(jid=AGENT_JID)

    await agent.start(auto_register=True)
    assert agent.is_alive() is True

    agent.result = None

    agent.pubsub.set_on_item_published(agent.callback)

    class NotifyBehaviour(OneShotBehaviour):
        async def run(self):
            await self.agent.pubsub.create(PUBSUB_JID, TEST_NODE)
            await self.agent.pubsub.subscribe(PUBSUB_JID, TEST_NODE)
            await self.agent.pubsub.notify(PUBSUB_JID, TEST_NODE)
            await self.agent.pubsub.delete(PUBSUB_JID, TEST_NODE)

    behaviour = NotifyBehaviour()
    agent.add_behaviour(behaviour)
    await behaviour.join()

    await agent.stop()
    assert agent.is_alive() is False


@pytest.mark.asyncio
async def test_publish_item(server):
    agent = PubSubAgentFactory(jid=AGENT_JID)

    await agent.start(auto_register=True)
    assert agent.is_alive() is True

    agent.pubsub.set_on_item_published(agent.callback)

    class PublishBehaviour(OneShotBehaviour):
        async def run(self):
            await self.agent.pubsub.create(PUBSUB_JID, TEST_NODE)
            await self.agent.pubsub.subscribe(PUBSUB_JID, TEST_NODE)
            await self.agent.pubsub.publish(PUBSUB_JID, TEST_NODE, TEST_PAYLOAD)
            items = await self.agent.pubsub.get_items(PUBSUB_JID, TEST_NODE)
            await self.agent.pubsub.delete(PUBSUB_JID, TEST_NODE)
            self.kill(exit_code=items)

    behaviour = PublishBehaviour()
    agent.add_behaviour(behaviour)
    await behaviour.join()

    assert len(behaviour.exit_code) == 1
    assert behaviour.exit_code[0][1].text == TEST_PAYLOAD

    await agent.stop()
    assert agent.is_alive() is False


@pytest.mark.asyncio
async def test_retract_item(server):
    agent = PubSubAgentFactory(jid=AGENT_JID)

    await agent.start(auto_register=True)
    assert agent.is_alive() is True

    agent.client.register_plugin('xep_0004')
    config_form = agent.client.plugin["xep_0004"].make_form(ftype="submit")
    config_form.addField('pubsub#persist_items', value=True)

    class PublishBehaviour(OneShotBehaviour):
        async def run(self):
            await self.agent.pubsub.create(PUBSUB_JID, TEST_NODE, config_form)
            await self.agent.pubsub.subscribe(PUBSUB_JID, TEST_NODE)
            await self.agent.pubsub.publish(PUBSUB_JID, TEST_NODE, TEST_PAYLOAD, ITEM_ID)
            items = await self.agent.pubsub.get_items(PUBSUB_JID, TEST_NODE)
            self.kill(exit_code=(items))

    publish_behaviour = PublishBehaviour()
    agent.add_behaviour(publish_behaviour)
    await publish_behaviour.join()

    assert len(publish_behaviour.exit_code) == 1
    assert TEST_PAYLOAD == publish_behaviour.exit_code[0][1].text

    class RetractBehaviour(OneShotBehaviour):
        async def run(self):
            await self.agent.pubsub.retract(PUBSUB_JID, TEST_NODE, ITEM_ID)
            items = await self.agent.pubsub.get_items(PUBSUB_JID, TEST_NODE)
            await self.agent.pubsub.delete(PUBSUB_JID, TEST_NODE)
            self.kill(exit_code=items)

    retract_behaviour = RetractBehaviour()
    agent.add_behaviour(retract_behaviour)
    await retract_behaviour.join()
    assert len(retract_behaviour.exit_code) == 0

    await agent.stop()
    assert agent.is_alive() is False



import factory
import slixmpp.stanza
from spade.agent import Agent

from spade_pubsub import PubSubMixin


class PubSubAgent(PubSubMixin, Agent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.result = None

    def callback(self, message: slixmpp.stanza.Message):
        self.result = message.get_payload()


class PubSubAgentFactory(factory.Factory):
    class Meta:
        model = PubSubAgent

    jid = "fake@jid"
    password = "fake_password"



import asyncio
import pytest
import pytest_asyncio

from loguru import logger
from pyjabber.server import Server, Parameters


@pytest.fixture(scope="module", autouse=True)
def cleanup(request):
    pass


@pytest_asyncio.fixture
async def server():
    logger.remove()

    server = Server(Parameters(database_in_memory=True))

    task = asyncio.create_task(server.start())
    yield
    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass



from loguru import logger
from typing import Optional, List, Union
from xml.etree.ElementTree import Element

from slixmpp import ClientXMPP
from slixmpp.exceptions import IqError
from slixmpp.plugins.xep_0004.stanza.form import Form
from slixmpp.plugins.xep_0060 import XEP_0060
from slixmpp.stanza import Iq


class PubSubMixin:
    """
    This mixin provides PubSub support to SPADE agents.
    It must be used as superclass of a spade.Agent subclass.
    """

    async def _hook_plugin_after_connection(self, *args, **kwargs):
        try:
            await super()._hook_plugin_after_connection(*args, **kwargs)
        except AttributeError:
            logger.debug("_hook_plugin_after_connection is undefined")

        self.pubsub = self.PubSubComponent(self.client)

    async def _hook_plugin_before_connection(self, *args, **kwargs):
        """
        Overload this method to hook a plugin before connetion is done
        """
        try:
            await super()._hook_plugin_before_connection(*args, **kwargs)
        except AttributeError:
            logger.debug("_hook_plugin_before_connection is undefined")

    class PubSubComponent:
        def __init__(self, client):
            self.client: ClientXMPP = client
            self.client.register_plugin("xep_0060")
            self.pubsub: XEP_0060 = self.client["xep_0060"]  # Pubsub XEP
            self.client.register_plugin("xep_0004")  # Dataforms XEP

        # OWNER USE CASES
        async def create(
            self,
            target_jid: str,
            target_node: Optional[str] = None,
            config_form: Optional[Form] = None,
        ):
            """
            Create a new node at a service.

            Args:
                target_jid (str): Address of the PubSub service.
                target_node (str or None): Name of the PubSub node to create
                config_form (Slixmpp Form): Dataform to configurate the node to create
            """
            try:
                res = await self.pubsub.create_node(
                    target_jid, target_node, config=config_form
                )
                if (
                    target_node
                    and res["pubsub"]
                    and res["pubsub"]["create"]
                    and res["pubsub"]["create"]["node"]
                ):
                    return res["pubsub"]["create"]["node"]
                else:
                    return target_node
            except IqError as e:
                logger.error(f"Error creating node <{target_node}>: {e}")

        async def delete(
            self,
            target_jid: str,
            target_node: Optional[str],
        ):
            """
            Delete an existing node.

            Args:
                target_jid (str): Address of the PubSub service.
                target_node (str or None): Name of the PubSub node to delete.
            """
            try:
                return await self.pubsub.delete_node(target_jid, target_node)
            except IqError as e:
                logger.error(f"Error deleting node <{target_node}>: {e}")

        async def get_node_subscriptions(
            self, target_jid: str, target_node: Optional[str]
        ) -> List[str]:
            """
            Return the subscriptions of other jids with a node.

            Args:
                target_jid (str): Address of the PubSub service.
                target_node (str): Name of the node to query
            """
            try:
                data: Iq = await self.pubsub.get_subscriptions(target_jid, target_node)
                if (
                    data["pubsub"]
                    and data["pubsub"]["subscriptions"]
                    and any(
                        i["node"] == target_node
                        for i in data["pubsub"]["subscriptions"]
                    )
                ):
                    return list(
                        set(
                            [
                                sub["jid"]
                                for sub in data["pubsub"]["subscriptions"]["substanzas"]
                            ]
                        )
                    )
                else:
                    return []
            except IqError as e:
                logger.error(
                    f"Error retrieving owner subscriptions from node <{target_node}>: {e}"
                )

        async def purge(self, target_jid: str, target_node: Optional[str]):
            """
            Delete all items from a node.

            Args:
                target_jid (str): JID of the PubSub service
                target_node (str): Name of the PubSub node
            """
            try:
                return await self.pubsub.purge(target_jid, target_node)
            except IqError as e:
                logger.error(f"Error purging node <{target_node}>: {e}")

        async def get_nodes(self, target_jid: str, target_node: Optional[str] = None):
            """
            Request all nodes at a service or collection node.

            Args:
                target_jid (str): Address of the PubSub service.
                target_node (str or None): Name of the collection node to query
            """
            try:
                nodes = await self.pubsub.get_nodes(target_jid, target_node)
                return [
                    {"jid": i[0], "node": i[1], "name": i[2]}
                    for i in nodes["disco_items"]["items"]
                ]
            except IqError as e:
                logger.error(f"Error retrieving nodes: {e}")
                return []

        async def get_items(
            self, target_jid: str, target_node: Optional[str]
        ) -> List[tuple[str, str]]:
            """
            Request all items at a service or collection node.

            Returns a list of tuples, in the format (id, payload)
            Args:
                target_jid (str): Address of the PubSub service.
                target_node (str or None): Name of the PubSub node.

            """
            try:
                data: Iq = await self.pubsub.get_items(target_jid, target_node)
                if data["pubsub"] and data:
                    if data["pubsub"]["items"]["node"] == target_node:
                        return [
                            (item["id"], item["payload"])
                            for item in data["pubsub"]["items"]
                        ]
            except IqError as e:
                logger.error(f"Error retrieving items from node <{target_node}>: {e}")

        # SUBSCRIBER USE CASES

        async def subscribe(
            self,
            target_jid: str,
            target_node: Optional[str] = None,
            subscription_jid: Optional[str] = None,
            config=None,
        ):
            """
            Subscribe to a node.

            Args:
                target_jid (str): Address of the PubSub service.
                target_node (str): Name of the PubSub node to subscribe to.
                subscription_jid (str): The address to subscribe to the service.
                config (Data): Optional configuration of the subscription
            """
            try:
                sub = await self.pubsub.subscribe(
                    target_jid, target_node, subscribee=subscription_jid, options=config
                )
                if sub["pubsub"] and sub["pubsub"]["subscription"]:
                    return sub["pubsub"]["subscription"]["subid"]
            except IqError as e:
                logger.error(f"Error subscribing to node <{target_node}>: {e}")

        async def unsubscribe(
            self,
            target_jid: str,
            target_node: Optional[str] = None,
            subscription_jid: Optional[str] = None,
            subid=None,
        ):
            """
            Unsubscribe from a node.

            Args:
                target_jid (str): Address of the PubSub service.
                target_node (str): Name of the PubSub node to unsubscribe from.
                subscription_jid (str): The address to subscribe from the service.
                subid (str): Unique ID of the subscription to remove.
            """
            try:
                return await self.pubsub.unsubscribe(
                    target_jid,
                    target_node,
                    subscribee=subscription_jid,
                    subid=subid,
                )
            except IqError as e:
                logger.error(f"Error unsubscribing to node <{target_node}>: {e}")

        def set_on_item_published(self, callback):
            self.client.add_event_handler("pubsub_publish", callback)

        def set_on_item_retracted(self, callback):
            self.client.add_event_handler("pubsub_retract", callback)

        # PUBLISHER USE CASES

        async def notify(self, target_jid: str, target_node: str):
            """
            Notify all subscribers of a node without publishing an item.
            “Publish” to the node at jid without any item. This merely fans out a notification.

            Args:
                target_jid (str): Address of the PubSub service.
                target_node (str): Name of the PubSub node to send a notify from.
            """
            try:
                await self.pubsub.publish(target_jid, target_node)
            except IqError as e:
                logger.error(f"Error notifying to node <{target_node}>: {e}")

        async def publish(
            self,
            target_jid: str,
            target_node: str,
            payload: Union[Element, str],
            item_id: Optional[str] = None,
            ifrom: str = None,
        ) -> Union[str, Iq]:
            """
            Publish an item to a node.

            Args:
                target_jid (str): Address of the PubSub service.
                target_node (str): Name of the PubSub node to publish to.
                payload (Element | str): Payload to publish.
                item_id (str or None): Item ID to use for the item.

            Return:
                The response of the server
            """
            try:
                payload_stanza = Element("payload", attrib={"xmlns": "spade.pubsub"})
                payload_stanza.text = payload

                res = await self.pubsub.publish(
                    target_jid, target_node, item_id, payload_stanza, ifrom=ifrom
                )
                if (
                    item_id is None
                    and res["pubsub"]
                    and res["pubsub"]["publish"]
                    and res["pubsub"]["publish"]["item"]
                ):
                    return res["pubsub"]["publish"]["item"]["id"]
            except IqError as e:
                logger.error(
                    f"Error publishing item <{item_id or 'undefined'}> to node <{target_node}>: {e}"
                )

        async def retract(
            self, target_jid: str, target_node: str, item_id: str, notify=False
        ):
            """
            Retract a previously published item from a node.

            Args:
                target_jid (str): Address of the PubSub service.
                target_node (str): Name of the PubSub node to send a notify from.
                item_id (str): The ID of the item to retract.
                notify (bool): Flag indicating whether subscribers shall be notified about the retraction.
            """
            try:
                return await self.pubsub.retract(
                    target_jid, target_node, item_id, notify
                )
            except IqError as e:
                logger.error(
                    f"Error retracting item <{item_id}> to node <{target_node}>: {e}"
                )



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = parse_requirements("requirements.txt")

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Javi Palanca",
    author_email='jpalanca@dsic.upv.es',
    python_requires='>=3.9',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    description="SPADE Plugin for PubSub support.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='spade_pubsub',
    name='spade_pubsub',
    packages=find_packages(include=['spade_pubsub', 'spade_pubsub.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/javipalanca/spade_pubsub',
    version="0.3.0",
    zip_safe=False,
)



"""Top-level package for SPADE PubSub."""

__author__ = """Javi Palanca"""
__email__ = "jpalanca@dsic.upv.es"
__version__ = "0.3.0"


from .pubsub import PubSubMixin

__all__ = ["PubSubMixin", "__version__"]
